import asyncio
import signal
import time
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path

from lite_dist2.common import float2hex, int2hex
from lite_dist2.config import TableConfig, TableConfigProvider, WorkerConfig
from lite_dist2.curriculum_models.study_portables import StudyRegistry
from lite_dist2.study_strategies import StudyStrategyModel
from lite_dist2.suggest_strategies import SuggestStrategyModel
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyParam
from lite_dist2.table_node_api.start_table_api import start_in_thread
from lite_dist2.table_node_api.table_param import StudyRegisterParam
from lite_dist2.type_definitions import RawParamType, RawResultType
from lite_dist2.value_models.aligned_space_registry import LineSegmentRegistry, ParameterAlignedSpaceRegistry
from lite_dist2.value_models.const_param import ConstParam, ConstParamElement
from lite_dist2.worker_node.table_node_client import TableNodeClient
from lite_dist2.worker_node.trial_runner import SemiAutoMPTrialRunner
from lite_dist2.worker_node.worker import Worker


def worker_init() -> None:
    # Ignore KeyboardInterrupt on worker processes
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def set_table_config() -> None:
    config = TableConfig(
        port=8000,
        curriculum_path=Path(__file__).parent / "example_curriculum.json",
        curriculum_save_interval_seconds=10,
    )
    TableConfigProvider.set(config)


async def register_study(table_ip: str, table_port: int) -> None:
    _resolution = 10
    _half_size = 2.0

    study_register_param = StudyRegisterParam(
        study=StudyRegistry(
            name=f"mandelbrot-{int(time.time())}",
            required_capacity=set(),
            study_strategy=StudyStrategyModel(type="all_calculation", study_strategy_param=None),
            suggest_strategy=SuggestStrategyModel(
                type="sequential",
                suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
            ),
            result_type="scalar",
            result_value_type="int",
            const_param=ConstParam(
                consts=[
                    ConstParamElement(type="float", key="abs_threshold", value=float2hex(2.0)),
                    ConstParamElement(type="int", key="max_iter", value=int2hex(255)),
                ],
            ),
            parameter_space=ParameterAlignedSpaceRegistry(
                type="aligned",
                axes=[
                    LineSegmentRegistry(
                        name="x",
                        type="float",
                        size=int2hex(_resolution),
                        step=float2hex(2 * _half_size / _resolution),
                        start=float2hex(-1 * _half_size),
                    ),
                    LineSegmentRegistry(
                        name="y",
                        type="float",
                        size=int2hex(_resolution),
                        step=float2hex(2 * _half_size / _resolution),
                        start=float2hex(-1 * _half_size),
                    ),
                ],
            ),
        ),
    )
    client = TableNodeClient(table_ip, table_port)
    await client.register_study(study_register_param)


class Mandelbrot(SemiAutoMPTrialRunner):
    def func(self, parameters: RawParamType, *_: object, **kwargs: object) -> RawResultType:
        abs_threshold = self.get_typed("abs_threshold", float, kwargs)
        max_iter = self.get_typed("max_iter", int, kwargs)
        x = float(parameters[0])
        y = float(parameters[1])
        c = complex(x, y)
        z = complex(0, 0)
        iter_count = 0
        while abs(z) <= abs_threshold and iter_count < max_iter:
            z = z**2 + c
            iter_count += 1
        return iter_count


async def run_worker(table_ip: str, table_port: int) -> None:
    worker_config = WorkerConfig(
        name="w_01",
        process_num=2,
        max_size=10,
        wait_seconds_on_no_trial=5,
        table_node_request_timeout_seconds=60,
    )
    with ProcessPoolExecutor(max_workers=2, initializer=worker_init) as pool:
        worker = Worker(
            trial_runner=Mandelbrot(),
            ip=table_ip,
            port=table_port,
            pool=pool,
            config=worker_config,
        )
        try:
            await worker.start_async(stop_at_no_trial=True)
        except KeyboardInterrupt:
            pool.shutdown(wait=False, cancel_futures=True)


async def main() -> None:
    # 1. Set table config for example
    #    Execution node type in normal use: Table
    #    NOTE: For normal use, change the `table_config.json` file in the root directory.
    set_table_config()

    # 2. Start table node in other thread
    #    Execution node type in normal use: Table
    #    NOTE: In most cases, use `start()` (or `uv run start`) instead of `start_in_thread()`
    #          since table nodes and worker nodes are separated.
    start_in_thread()
    await asyncio.sleep(1)  # wait for activate

    # 3. Register study to table node
    #    Execution node type in normal use: Management
    #    NOTE: If the management node is also a worker node or a table node,
    #          `TableNodeClient.register_study` is available.
    #          Otherwise, studies can be registered using the curl command.
    await register_study(table_ip="127.0.0.1", table_port=8000)

    # 4. run worker node
    #    Execution node type in normal use: Worker
    #    NOTE: In this example, the entire process should be completed within 10 seconds.
    #          10 seconds later, example_curriculum.json should be saved in the same directory as this file.
    #          The saving interval can be changed from the `TableConfig.curriculum_save_interval_seconds`.
    await run_worker(table_ip="127.0.0.1", table_port=8000)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from typing import TYPE_CHECKING, override

from lite_dist2.curriculum_models.mapping import MappingsStorage
from lite_dist2.curriculum_models.trial import Trial
from lite_dist2.expections import LD2NotDoneError
from lite_dist2.study_strategies import BaseStudyStrategy, StudyStrategyModel

if TYPE_CHECKING:
    from lite_dist2.curriculum_models.mapping import Mapping
    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.study_strategies.base_study_strategy import StudyStrategyParam
    from lite_dist2.trial_repositories.base_trial_repository import BaseTrialRepository
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace


class FindExactStudyStrategy(BaseStudyStrategy):
    def __init__(self, study_strategy_param: StudyStrategyParam) -> None:
        self.found_mapping: Mapping | None = None
        self.study_strategy_param = study_strategy_param

    @override
    async def is_done(
        self,
        trial_table: TrialTable,
        parameter_space: ParameterAlignedSpace,
        trial_repository: BaseTrialRepository,
    ) -> bool:
        if self.found_mapping:
            return True
        self.found_mapping = await self._find(trial_repository)
        return bool(self.found_mapping)

    async def _find(self, trial_repository: BaseTrialRepository) -> Mapping | None:
        trials = await trial_repository.load_all()
        for trial in trials:
            finding = Trial.from_model(trial).find_target_value(self.study_strategy_param.target_value)
            if finding:
                return finding
        return None

    @override
    async def extract_mappings(self, trial_repository: BaseTrialRepository) -> MappingsStorage:
        if not self.found_mapping:
            self.found_mapping = await self._find(trial_repository)
        if not self.found_mapping:
            raise LD2NotDoneError

        params = tuple(param.to_dummy() for param in self.found_mapping.params)
        result = self.found_mapping.result.to_dummy()

        return MappingsStorage(params_info=params, result_info=result, values=[self.found_mapping.to_tuple()])

    @override
    def to_model(self) -> StudyStrategyModel:
        return StudyStrategyModel(
            type="find_exact",
            study_strategy_param=self.study_strategy_param,
        )



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

if TYPE_CHECKING:
    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.value_models.space_type import ParameterSpaceType


class SuggestStrategyParam(BaseModel):
    strict_aligned: bool


class SuggestStrategyModel(BaseModel):
    type: Literal["sequential", "random", "designated"]
    suggest_strategy_param: SuggestStrategyParam


class BaseSuggestStrategy(abc.ABC):
    @abc.abstractmethod
    def to_model(self) -> SuggestStrategyModel:
        pass

    @abc.abstractmethod
    def suggest(self, trial_table: TrialTable, max_num: int) -> ParameterSpaceType | None:
        pass



from lite_dist2.value_models.aligned_space import ParameterAlignedSpace
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace

type ParameterSpaceType = ParameterAlignedSpace | ParameterJaggedSpace



from __future__ import annotations

from typing import TYPE_CHECKING

from lite_dist2.expections import LD2ModelTypeError, LD2ParameterError
from lite_dist2.study_strategies.all_calculation_study_strategy import AllCalculationStudyStrategy
from lite_dist2.study_strategies.find_exact_study_strategy import FindExactStudyStrategy

if TYPE_CHECKING:
    from lite_dist2.study_strategies import BaseStudyStrategy, StudyStrategyModel


def create_study_strategy(model: StudyStrategyModel) -> BaseStudyStrategy:
    param = model.study_strategy_param
    match model.type:
        case "all_calculation":
            return AllCalculationStudyStrategy(param)
        case "find_exact":
            if param is None:
                p = "study_strategy_param"
                et = "missing"
                raise LD2ParameterError(p, et)
            return FindExactStudyStrategy(param)
        case "minimize":
            raise NotImplementedError
        case _:
            raise LD2ModelTypeError(model.type)



from pathlib import Path

from pydantic import BaseModel

from lite_dist2.type_definitions import TrialRepositoryType


class TrialRepositoryModel(BaseModel):
    type: TrialRepositoryType
    save_dir: Path



from __future__ import annotations

from itertools import starmap
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.common import float2hex, int2hex, numerize
from lite_dist2.expections import LD2UndefinedError

if TYPE_CHECKING:
    from lite_dist2.type_definitions import ConstParamType


class ConstParamElement(BaseModel):
    type: Literal["int", "float", "bool", "str"]
    key: str
    value: bool | str

    def unpack(self) -> ConstParamType:
        match self.type:
            case "int" | "float" | "bool":
                return numerize(self.type, self.value)
            case "str":
                return self.value
            case _:
                raise LD2UndefinedError(self.type)

    @staticmethod
    def from_kv(key: str, value: ConstParamType) -> ConstParamElement:
        match value:
            case bool():
                return ConstParamElement(type="bool", key=key, value=value)
            case int():
                return ConstParamElement(type="int", key=key, value=int2hex(value))
            case float():
                return ConstParamElement(type="float", key=key, value=float2hex(value))
            case str():
                return ConstParamElement(type="str", key=key, value=value)
            case _:
                raise LD2UndefinedError(type(value))


class ConstParam(BaseModel):
    consts: list[ConstParamElement]

    def to_dict(self) -> dict[str, ConstParamType]:
        return {const.key: const.unpack() for const in self.consts}

    @staticmethod
    def from_dict(d: dict[str, ConstParamType]) -> ConstParam:
        return ConstParam(consts=list(starmap(ConstParamElement.from_kv, d.items())))



from datetime import datetime

import pytest

from lite_dist2.curriculum_models.mapping import Mapping
from lite_dist2.curriculum_models.trial import Trial, TrialDoneRecord, TrialModel, TrialStatus
from lite_dist2.expections import LD2NotDoneError
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpacePortableModel
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace, ParameterJaggedSpacePortableModel
from lite_dist2.value_models.line_segment import DummyLineSegment, LineSegment, LineSegmentPortableModel
from lite_dist2.value_models.point import ResultType, ScalarValue, VectorValue
from tests.const import DT, JST

_dummy_space = ParameterJaggedSpace(
    parameters=[(0, 1)],
    ambient_indices=[(0, 1)],
    axes_info=[DummyLineSegment(name="x", type_="int", ambient_size=6, step=1)],
)

_dummy_aligned_space = ParameterAlignedSpace(
    axes=[LineSegment(name="x", type_="int", size=2, step=1, start=0, ambient_index=1, ambient_size=10)],
    check_lower_filling=True,
)


def test_trial_done_record_calc_duration_sec() -> None:
    rec = TrialDoneRecord(
        trial_id="t01",
        reserved_timestamp=datetime(2025, 4, 27, 20, 36, 10, 0, tzinfo=JST),
        worker_node_name="w01",
        worker_node_id="w01",
        registered_timestamp=datetime(2025, 4, 27, 20, 36, 11, 500_000, tzinfo=JST),
        grid_size=120,
    )
    actual = rec.calc_duration_sec()
    expected = 1.5
    assert actual == expected


def test_trial_done_record_calc_grid_per_sec() -> None:
    rec = TrialDoneRecord(
        trial_id="t01",
        reserved_timestamp=datetime(2025, 4, 27, 20, 36, 10, 0, tzinfo=JST),
        worker_node_name="w01",
        worker_node_id="w01",
        registered_timestamp=datetime(2025, 4, 27, 20, 36, 11, 500_000, tzinfo=JST),
        grid_size=120,
    )
    actual = rec.calc_grid_per_sec()
    expected = 80.0
    assert actual == expected


@pytest.mark.parametrize(
    ("trial", "target", "expected"),
    [
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                reserved_timestamp=DT,
                trial_status=TrialStatus.running,
                const_param=None,
                parameter_space=_dummy_space,
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                        result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                    ),
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x1"),),
                        result=ScalarValue(type="scalar", value_type="int", value="0x67"),
                    ),
                ],
            ),
            ScalarValue(type="scalar", value_type="int", value="0x66"),
            None,
            id="not found: running",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                reserved_timestamp=DT,
                trial_status=TrialStatus.done,
                const_param=None,
                parameter_space=_dummy_space,
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=None,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x66"),
            None,
            id="not found: none result",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                reserved_timestamp=DT,
                trial_status=TrialStatus.done,
                const_param=None,
                parameter_space=_dummy_space,
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[],
            ),
            ScalarValue(type="scalar", value_type="int", value="0x66"),
            None,
            id="not found: empty result",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                reserved_timestamp=DT,
                trial_status=TrialStatus.done,
                const_param=None,
                parameter_space=_dummy_space,
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                        result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                    ),
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x1"),),
                        result=ScalarValue(type="scalar", value_type="int", value="0x67"),
                    ),
                ],
            ),
            ScalarValue(type="scalar", value_type="int", value="0x66"),
            Mapping(
                params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
            ),
            id="found: scalar",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                reserved_timestamp=DT,
                trial_status=TrialStatus.done,
                const_param=None,
                parameter_space=_dummy_space,
                result_type="vector",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                        result=VectorValue(type="vector", value_type="int", values=["0x66", "0x2328"]),
                    ),
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x1"),),
                        result=VectorValue(type="vector", value_type="int", values=["0x67", "0x2329"]),
                    ),
                ],
            ),
            VectorValue(type="vector", value_type="int", values=["0x66", "0x2328"]),
            Mapping(
                params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                result=VectorValue(type="vector", value_type="int", values=["0x66", "0x2328"]),
            ),
            id="found: vector",
        ),
    ],
)
def test_trial_find_target_value(trial: Trial, target: ResultType, expected: Mapping | None) -> None:
    actual = trial.find_target_value(target)
    assert actual == expected


def test_to_done_record() -> None:
    trial = Trial(
        study_id="s01",
        trial_id="t02",
        reserved_timestamp=datetime(2025, 4, 27, 20, 36, 10, 0, tzinfo=JST),
        trial_status=TrialStatus.done,
        const_param=None,
        parameter_space=_dummy_aligned_space,
        result_type="scalar",
        result_value_type="float",
        worker_node_name="w01",
        worker_node_id="w01",
        results=[
            Mapping(
                params=(
                    ScalarValue(
                        type="scalar",
                        value_type="int",
                        value="0x1",
                        name="x",
                    ),
                ),
                result=ScalarValue(
                    type="scalar",
                    value_type="float",
                    value="0x1.0000000000000p+1",
                    name="r1",
                ),
            ),
            Mapping(
                params=(
                    ScalarValue(
                        type="scalar",
                        value_type="int",
                        value="0x2",
                        name="x",
                    ),
                ),
                result=ScalarValue(
                    type="scalar",
                    value_type="float",
                    value="0x1.0000000000000p+2",
                    name="r1",
                ),
            ),
        ],
        registered_timestamp=datetime(2025, 4, 27, 20, 36, 11, 500, tzinfo=JST),
    )
    expected = TrialDoneRecord(
        trial_id="t02",
        reserved_timestamp=datetime(2025, 4, 27, 20, 36, 10, 0, tzinfo=JST),
        worker_node_name="w01",
        worker_node_id="w01",
        registered_timestamp=datetime(2025, 4, 27, 20, 36, 11, 500, tzinfo=JST),
        grid_size=2,
    )

    actual = trial.to_done_record()
    assert actual == expected


def test_to_done_record_raises() -> None:
    trial = Trial(
        study_id="s01",
        trial_id="t02",
        reserved_timestamp=datetime(2025, 4, 27, 20, 36, 10, 0, tzinfo=JST),
        trial_status=TrialStatus.running,
        const_param=None,
        parameter_space=_dummy_aligned_space,
        result_type="scalar",
        result_value_type="float",
        worker_node_name="w01",
        worker_node_id="w01",
        results=None,
        registered_timestamp=None,
    )

    with pytest.raises(LD2NotDoneError):
        _ = trial.to_done_record()


@pytest.mark.parametrize(
    ("trial", "cutoff_datetime", "expected"),
    [
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t02",
                reserved_timestamp=DT,
                trial_status=TrialStatus.running,
                const_param=None,
                parameter_space=_dummy_aligned_space,
                result_type="scalar",
                result_value_type="float",
                worker_node_name="w01",
                worker_node_id="w01",
                results=None,
                registered_timestamp=None,
            ),
            datetime(2025, 4, 27, 20, 36, 10, 0, tzinfo=JST),
            False,
            id="Not done (status, registered_datetime): False",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t02",
                reserved_timestamp=DT,
                trial_status=TrialStatus.done,
                const_param=None,
                parameter_space=_dummy_aligned_space,
                result_type="scalar",
                result_value_type="float",
                worker_node_name="w01",
                worker_node_id="w01",
                results=None,
                registered_timestamp=None,
            ),
            datetime(2025, 4, 27, 20, 36, 10, 0, tzinfo=JST),
            False,
            id="Not done (registered_datetime): False",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t02",
                reserved_timestamp=DT,
                trial_status=TrialStatus.running,
                const_param=None,
                parameter_space=_dummy_aligned_space,
                result_type="scalar",
                result_value_type="float",
                worker_node_name="w01",
                worker_node_id="w01",
                results=None,
                registered_timestamp=DT,
            ),
            datetime(2025, 4, 27, 20, 36, 9, 0, tzinfo=JST),
            False,
            id="Not done (status): False",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t02",
                reserved_timestamp=DT,
                trial_status=TrialStatus.done,
                const_param=None,
                parameter_space=_dummy_aligned_space,
                result_type="scalar",
                result_value_type="float",
                worker_node_name="w01",
                worker_node_id="w01",
                results=None,
                registered_timestamp=datetime(2025, 4, 27, 20, 36, 11, 0, tzinfo=JST),
            ),
            datetime(2025, 4, 27, 20, 36, 12, 0, tzinfo=JST),
            False,
            id="Done, but past: False",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t02",
                reserved_timestamp=DT,
                trial_status=TrialStatus.done,
                const_param=None,
                parameter_space=_dummy_aligned_space,
                result_type="scalar",
                result_value_type="float",
                worker_node_name="w01",
                worker_node_id="w01",
                results=None,
                registered_timestamp=datetime(2025, 4, 27, 20, 36, 12, 0, tzinfo=JST),
            ),
            datetime(2025, 4, 27, 20, 36, 11, 0, tzinfo=JST),
            True,
            id="Done, after cutoff datetime: True",
        ),
    ],
)
def test_done_in_after(trial: Trial, cutoff_datetime: datetime, expected: bool) -> None:
    actual = trial.done_in_after(cutoff_datetime)
    assert actual == expected


@pytest.mark.parametrize(
    "model",
    [
        TrialModel(
            study_id="my_study_id0",
            trial_id="01",
            reserved_timestamp=DT,
            trial_status=TrialStatus.done,
            const_param=None,
            parameter_space=ParameterAlignedSpacePortableModel(
                type="aligned",
                axes=[
                    LineSegmentPortableModel(
                        name="x",
                        type="bool",
                        size="0x2",
                        step=True,
                        start=False,
                        ambient_index="0x0",
                        ambient_size="0x2",
                    ),
                ],
                check_lower_filling=True,
            ),
            result_type="scalar",
            result_value_type="bool",
            worker_node_name="w01",
            worker_node_id="w01",
        ),
        TrialModel(
            study_id="my_study_id1",
            trial_id="01",
            reserved_timestamp=DT,
            trial_status=TrialStatus.done,
            const_param=None,
            parameter_space=ParameterAlignedSpacePortableModel(
                type="aligned",
                axes=[
                    LineSegmentPortableModel(
                        name="x",
                        type="int",
                        size="0x2",
                        step="0x1",
                        start="0x0",
                        ambient_index="0x0",
                        ambient_size="0x2",
                    ),
                ],
                check_lower_filling=True,
            ),
            result_type="scalar",
            result_value_type="float",
            worker_node_name="w01",
            worker_node_id="w01",
            results=[
                Mapping(
                    params=(
                        ScalarValue(
                            type="scalar",
                            value_type="int",
                            value="0x1",
                            name="x",
                        ),
                    ),
                    result=ScalarValue(
                        type="scalar",
                        value_type="float",
                        value="0x1.0000000000000p+1",
                        name="r1",
                    ),
                ),
            ],
        ),
        TrialModel(
            study_id="my_study_id2",
            trial_id="02",
            reserved_timestamp=DT,
            trial_status=TrialStatus.done,
            const_param=None,
            parameter_space=ParameterJaggedSpacePortableModel(
                type="jagged",
                axes_info=[
                    LineSegmentPortableModel(
                        name="x",
                        type="int",
                        size="0x1",
                        start="0x0",
                        step="0x1",
                        ambient_index="0x0",
                        ambient_size="0x1",
                        is_dummy=True,
                    ),
                ],
                parameters=[("0x1",)],
                ambient_indices=[("0x1",)],
            ),
            result_type="scalar",
            result_value_type="int",
            worker_node_name="w01",
            worker_node_id="w01",
        ),
        TrialModel(
            study_id="my_study_id3",
            trial_id="01",
            reserved_timestamp=DT,
            trial_status=TrialStatus.done,
            const_param=None,
            parameter_space=ParameterJaggedSpacePortableModel(
                type="jagged",
                axes_info=[
                    LineSegmentPortableModel(
                        name="x",
                        type="int",
                        size="0x1",
                        start="0x0",
                        step="0x1",
                        ambient_index="0x0",
                        ambient_size="0x1",
                        is_dummy=True,
                    ),
                ],
                parameters=[("0x1",)],
                ambient_indices=[("0x1",)],
            ),
            result_type="scalar",
            result_value_type="int",
            worker_node_name="w01",
            worker_node_id="w01",
            results=[
                Mapping(
                    params=(
                        ScalarValue(
                            type="scalar",
                            value_type="int",
                            value="0x1",
                            name="x",
                        ),
                    ),
                    result=ScalarValue(
                        type="scalar",
                        value_type="int",
                        value="0x2",
                        name="r1",
                    ),
                ),
            ],
        ),
    ],
)
def test_trial_to_model_from_model(model: TrialModel) -> None:
    trial = Trial.from_model(model)
    reconstructed_trial = trial.to_model()
    assert model == reconstructed_trial



from __future__ import annotations

import functools
from typing import Literal

from pydantic import BaseModel

from lite_dist2.common import hex2int
from lite_dist2.type_definitions import PortableValueType
from lite_dist2.value_models.line_segment import LineSegmentPortableModel


class ParameterAlignedSpacePortableModel(BaseModel):
    type: Literal["aligned"]
    axes: list[LineSegmentPortableModel]
    check_lower_filling: bool

    @functools.cached_property
    def total(self) -> int | None:
        t = 1
        for axis in self.axes:
            if axis.size is None:
                return None
            t *= hex2int(axis.size)
        return t


class ParameterJaggedSpacePortableModel(BaseModel):
    type: Literal["jagged"]
    parameters: list[tuple[PortableValueType, ...]]
    ambient_indices: list[tuple[str, ...]]
    axes_info: list[LineSegmentPortableModel]


type SpacePortableModelType = ParameterAlignedSpacePortableModel | ParameterJaggedSpacePortableModel



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, override

import pytest

from lite_dist2.curriculum_models.mapping import Mapping, MappingsStorage
from lite_dist2.curriculum_models.trial import TrialModel, TrialStatus
from lite_dist2.curriculum_models.trial_table import TrialTable
from lite_dist2.expections import LD2NotDoneError
from lite_dist2.study_strategies.base_study_strategy import StudyStrategyParam
from lite_dist2.study_strategies.find_exact_study_strategy import FindExactStudyStrategy
from lite_dist2.trial_repositories.base_trial_repository import BaseTrialRepository
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpacePortableModel
from lite_dist2.value_models.line_segment import LineSegment, LineSegmentPortableModel
from lite_dist2.value_models.point import ResultType, ScalarValue
from tests.const import DT

if TYPE_CHECKING:
    from lite_dist2.trial_repositories.trial_repository_model import TrialRepositoryModel
    from lite_dist2.type_definitions import TrialRepositoryType

_DUMMY_PARAMETER_SPACE_MODEL = ParameterAlignedSpacePortableModel(
    type="aligned",
    axes=[
        LineSegmentPortableModel(
            name="x",
            type="int",
            size="0x2",
            start="0x0",
            ambient_size="0x2",
            ambient_index="0x0",
            step="0x1",
        ),
        LineSegmentPortableModel(
            name="y",
            type="int",
            size="0x2",
            start="0x0",
            ambient_size="0x2",
            ambient_index="0x0",
            step="0x1",
        ),
    ],
    check_lower_filling=True,
)


class MockTrialRepository(BaseTrialRepository):
    def __init__(self) -> None:
        self.save_dir = Path("test/s01")

    @override
    @staticmethod
    def get_repository_type() -> TrialRepositoryType:
        return "normal"

    @override
    async def clean_save_dir(self) -> None:
        pass

    @override
    async def save(self, trial: TrialModel) -> None:
        pass

    @override
    async def load(self, trial_id: str) -> TrialModel:
        raise NotImplementedError

    @override
    async def load_all(self) -> list[TrialModel]:
        raise NotImplementedError

    @override
    async def delete_save_dir(self) -> None:
        pass

    @override
    def to_model(self) -> TrialRepositoryModel:
        raise NotImplementedError


@pytest.fixture
def trial_repository_fixture(monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest) -> None:
    return_value = request.param

    async def fake_load_all(self) -> int:  # noqa: ANN001
        return return_value

    monkeypatch.setattr(MockTrialRepository, "load_all", fake_load_all)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("trial_repository_fixture", "target_value", "expected"),
    [
        pytest.param(
            [],
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            False,
            id="Empty: False",
        ),
        pytest.param(
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            False,
            id="Not found: False",
        ),
        pytest.param(
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            True,
            id="Found: True",
        ),
        pytest.param(
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.running,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            False,
            id="Found but running: False",
        ),
    ],
    indirect=["trial_repository_fixture"],
)
async def test_find_exact_study_strategy_is_done(
    trial_repository_fixture: list[TrialModel],
    target_value: ResultType,
    expected: bool,
) -> None:
    table = TrialTable(trials=[], aggregated_parameter_space=None)
    strategy = FindExactStudyStrategy(StudyStrategyParam(target_value=target_value))
    parameter_space = ParameterAlignedSpace(
        axes=[
            LineSegment(name="x", type_="int", size=2, step=1, start=0, ambient_size=2, ambient_index=0),
            LineSegment(name="y", type_="int", size=2, step=1, start=0, ambient_size=2, ambient_index=0),
        ],
        check_lower_filling=True,
    )

    actual = await strategy.is_done(table, parameter_space, MockTrialRepository())
    assert actual == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("target_value", "trial_repository_fixture", "expected"),
    [
        pytest.param(
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            MappingsStorage(
                params_info=(
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                ),
                result_info=ScalarValue(type="scalar", value_type="int", value="0x0"),
                values=[
                    ("0x0", "0x0", "0x64"),
                ],
            ),
            id="Found",
        ),
    ],
    indirect=["trial_repository_fixture"],
)
async def test_find_exact_study_strategy_extract_mapping(
    target_value: ResultType,
    trial_repository_fixture: list[TrialModel],
    expected: MappingsStorage,
) -> None:
    repo = MockTrialRepository()
    strategy = FindExactStudyStrategy(StudyStrategyParam(target_value=target_value))
    actual = await strategy.extract_mappings(repo)
    assert actual == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("target_value", "trial_repository_fixture"),
    [
        pytest.param(
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            [],
            id="Empty",
        ),
        pytest.param(
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            id="Not found",
        ),
    ],
    indirect=["trial_repository_fixture"],
)
async def test_find_exact_study_strategy_extract_mapping_raise(
    target_value: ResultType,
    trial_repository_fixture: list[TrialModel],
) -> None:
    repo = MockTrialRepository()
    strategy = FindExactStudyStrategy(StudyStrategyParam(target_value=target_value))
    with pytest.raises(LD2NotDoneError):
        _ = await strategy.extract_mappings(repo)



from __future__ import annotations

import abc
import functools
import logging
from concurrent.futures import ProcessPoolExecutor, as_completed
from multiprocessing.pool import Pool
from typing import TYPE_CHECKING, Any, override

import tqdm

from lite_dist2.expections import LD2TypeError
from lite_dist2.type_definitions import ConstParamType, PrimitiveValueType

if TYPE_CHECKING:
    from collections.abc import Iterator

    from lite_dist2.config import WorkerConfig
    from lite_dist2.curriculum_models.trial import Trial
    from lite_dist2.type_definitions import RawParamType, RawResultType
    from lite_dist2.value_models.space_type import ParameterSpaceType


logger = logging.getLogger(__name__)


class BaseTrialRunner(abc.ABC):
    @abc.abstractmethod
    def func(self, parameters: RawParamType, *args: object, **kwargs: object) -> RawResultType:
        pass

    @abc.abstractmethod
    def wrap_func(
        self,
        parameter_space: ParameterSpaceType,
        config: WorkerConfig,
        pool: Pool | ProcessPoolExecutor | None = None,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        pass

    def parameter_pass_func(
        self,
        parameters: RawParamType,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> tuple[RawParamType, RawResultType]:
        return parameters, self.func(parameters, *args, **kwargs)

    def run(
        self,
        trial: Trial,
        config: WorkerConfig,
        pool: Pool | ProcessPoolExecutor | None = None,
        *args: object,
        **kwargs: object,
    ) -> Trial:
        raw_mappings = self.wrap_func(trial.parameter_space, config, pool, *args, **kwargs)
        mappings = trial.convert_mappings_from(raw_mappings)
        trial.set_result(mappings)
        return trial

    @staticmethod
    def get_typed[T](key: str, value_type: type[T], d: dict[str, object]) -> T:
        v = d.get(key)
        if isinstance(v, value_type):
            return v
        raise LD2TypeError(key, ConstParamType.__value__, type(v))


class AutoMPTrialRunner(BaseTrialRunner, abc.ABC):
    @override
    def wrap_func(
        self,
        parameter_space: ParameterSpaceType,
        config: WorkerConfig,
        pool: Pool | ProcessPoolExecutor | None = None,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        raw_mappings: list[tuple[RawParamType, RawResultType]] = []
        total = parameter_space.total
        tqdm_kwargs = {"total": total, "disable": config.disable_function_progress_bar}
        if config.process_num is None or config.process_num > 1:
            parameter_pass_func = functools.partial(self.parameter_pass_func, args=args, kwargs=kwargs)
            _pool = Pool(processes=config.process_num)
            try:
                with tqdm.tqdm(**tqdm_kwargs) as p_bar:
                    for arg_tuple, result_iter in _pool.imap_unordered(
                        func=parameter_pass_func,
                        iterable=parameter_space.grid(),
                        chunksize=config.chunk_size,
                    ):
                        raw_mappings.append((arg_tuple, result_iter))
                        p_bar.update(1)
                return raw_mappings
            except KeyboardInterrupt:
                _pool.terminate()
                _pool.join()
            else:
                _pool.close()
                _pool.join()
        return [
            self.parameter_pass_func(arg_tuple, args, kwargs)
            for arg_tuple in tqdm.tqdm(parameter_space.grid(), **tqdm_kwargs)
        ]


class SemiAutoMPTrialRunner(BaseTrialRunner, abc.ABC):
    @override
    def wrap_func(
        self,
        parameter_space: ParameterSpaceType,
        config: WorkerConfig,
        pool: Pool | ProcessPoolExecutor | None = None,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        total = parameter_space.total
        tqdm_kwargs = {"total": total, "disable": config.disable_function_progress_bar}

        if pool is None:
            logger.warning("pool is None, running in single-threaded mode")
            return [
                self.parameter_pass_func(arg_tuple, args, kwargs)
                for arg_tuple in tqdm.tqdm(parameter_space.grid(), **tqdm_kwargs)
            ]

        parameter_pass_func = functools.partial(self.parameter_pass_func, args=args, kwargs=kwargs)
        grid = parameter_space.grid()
        match pool:
            case Pool():
                return self._run_pool(pool, parameter_pass_func, grid, config.chunk_size, tqdm_kwargs)
            case ProcessPoolExecutor():
                return self._run_process_pool_executor(pool, parameter_pass_func, grid, tqdm_kwargs)
            case _:
                n = "pool"
                raise LD2TypeError(n, "Pool or ProcessPoolExecutor", type(pool))

    def _run_pool(
        self,
        pool: Pool,
        parameter_pass_func: functools.partial[tuple[RawParamType, RawResultType]],
        grid: Iterator[tuple[PrimitiveValueType, ...]],
        chunk_size: int,
        tqdm_kwargs: dict[str, object],
    ) -> list[tuple[RawParamType, RawResultType]]:
        raw_mappings: list[tuple[RawParamType, RawResultType]] = []
        with tqdm.tqdm(**tqdm_kwargs) as p_bar:
            for arg_tuple, result_iter in pool.imap_unordered(
                func=parameter_pass_func,
                iterable=grid,
                chunksize=chunk_size,
            ):
                raw_mappings.append((arg_tuple, result_iter))
                p_bar.update(1)
        return raw_mappings

    def _run_process_pool_executor(
        self,
        pool: ProcessPoolExecutor,
        parameter_pass_func: functools.partial[tuple[RawParamType, RawResultType]],
        grid: Iterator[tuple[PrimitiveValueType, ...]],
        tqdm_kwargs: dict[str, object],
    ) -> list[tuple[RawParamType, RawResultType]]:
        raw_mappings: list[tuple[RawParamType, RawResultType]] = []
        futures = {pool.submit(parameter_pass_func, arg_tuple): arg_tuple for arg_tuple in grid}
        with tqdm.tqdm(**tqdm_kwargs) as p_bar:
            for future in as_completed(futures):
                arg_tuple, result_iter = future.result()
                raw_mappings.append((arg_tuple, result_iter))
                p_bar.update(1)
        return raw_mappings


class ManualMPTrialRunner(BaseTrialRunner, abc.ABC):
    @override
    def func(self, parameters: RawParamType, *args: object, **kwargs: object) -> RawResultType:
        return 0

    @abc.abstractmethod
    def batch_func(
        self,
        raw_params: Iterator[RawParamType],
        config: WorkerConfig,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        pass

    @override
    def wrap_func(
        self,
        parameter_space: ParameterSpaceType,
        config: WorkerConfig,
        pool: Pool | ProcessPoolExecutor | None = None,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        return self.batch_func(parameter_space.grid(), config, *args, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel

from lite_dist2.type_definitions import PortableValueType
from lite_dist2.value_models.point import ParamType, ResultType

if TYPE_CHECKING:
    from typing import Literal


class Mapping(BaseModel):
    params: ParamType
    result: ResultType

    def to_tuple(self) -> tuple[PortableValueType, ...]:
        values = [param.value for param in self.params]
        values.extend(self.result.get_value_list())
        return tuple(values)


class MappingsStorage(BaseModel):
    params_info: ParamType
    result_info: ResultType
    values: list[tuple[PortableValueType, ...]]

    def get_names(self) -> tuple[str | None, ...]:
        names = [param.name for param in self.params_info]
        names.append(self.result_info.name)
        return tuple(names)

    def are_results(self) -> tuple[bool, ...]:
        p_size = len(self.params_info)
        r_size = self.result_info.get_value_size()
        return tuple([False] * p_size + [True] * r_size)

    def get_types(self) -> tuple[Literal["bool", "int", "float"], ...]:
        types: list[Literal["bool", "int", "float"]] = [param.value_type for param in self.params_info]
        types.extend(self.result_info.get_value_types())
        return tuple(types)



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.value_models.point import ResultType

if TYPE_CHECKING:
    from lite_dist2.curriculum_models.mapping import MappingsStorage
    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.trial_repositories.base_trial_repository import BaseTrialRepository
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace


class StudyStrategyParam(BaseModel):
    target_value: ResultType


class StudyStrategyModel(BaseModel):
    type: Literal["all_calculation", "find_exact", "minimize"]
    study_strategy_param: StudyStrategyParam | None


class BaseStudyStrategy(abc.ABC):
    @abc.abstractmethod
    async def is_done(
        self,
        trial_table: TrialTable,
        parameter_space: ParameterAlignedSpace,
        trial_repository: BaseTrialRepository,
    ) -> bool:
        pass

    @abc.abstractmethod
    async def extract_mappings(self, trial_repository: BaseTrialRepository) -> MappingsStorage:
        pass

    @abc.abstractmethod
    def to_model(self) -> StudyStrategyModel:
        pass



import asyncio
import logging
import time

from lite_dist2.common import float2hex, int2hex
from lite_dist2.curriculum_models.study_portables import StudyRegistry
from lite_dist2.study_strategies import StudyStrategyModel
from lite_dist2.suggest_strategies import SuggestStrategyModel
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyParam
from lite_dist2.table_node_api.table_param import StudyRegisterParam
from lite_dist2.value_models.aligned_space_registry import LineSegmentRegistry, ParameterAlignedSpaceRegistry
from lite_dist2.value_models.const_param import ConstParam, ConstParamElement
from lite_dist2.worker_node.table_node_client import TableNodeClient

logger = logging.getLogger(__name__)

async def register_study(table_ip: str, table_port: int) -> None:
    _resolution = 10
    _half_size = 2.0

    study_register_param = StudyRegisterParam(
        study=StudyRegistry(
            name=f"mandelbrot-{int(time.time())}",
            required_capacity=set(),
            study_strategy=StudyStrategyModel(type="all_calculation", study_strategy_param=None),
            suggest_strategy=SuggestStrategyModel(
                type="sequential",
                suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
            ),
            result_type="scalar",
            result_value_type="int",
            const_param=ConstParam(
                consts=[
                    ConstParamElement(type="float", key="abs_threshold", value=float2hex(2.0)),
                    ConstParamElement(type="int", key="max_iter", value=int2hex(255)),
                ],
            ),
            parameter_space=ParameterAlignedSpaceRegistry(
                type="aligned",
                axes=[
                    LineSegmentRegistry(
                        name="x",
                        type="float",
                        size=int2hex(_resolution),
                        step=float2hex(2 * _half_size / _resolution),
                        start=float2hex(-1 * _half_size),
                    ),
                    LineSegmentRegistry(
                        name="y",
                        type="float",
                        size=int2hex(_resolution),
                        step=float2hex(2 * _half_size / _resolution),
                        start=float2hex(-1 * _half_size),
                    ),
                ],
            ),
        ),
    )
    client = TableNodeClient(table_ip, table_port)
    await client.register_study(study_register_param)
    logger.info("Study registered: %s", study_register_param.study.name)


async def main() -> None:
    # Wait for table node to be ready
    for i in range(30):
        try:
            await register_study(table_ip="table", table_port=8000)  # Docker Compose でのサービス名
            logger.info("Management: Study registration completed")
        except Exception as e:  # noqa: BLE001
            logger.warning("Waiting for table node... (%d/30): %s", i+1, e)
            await asyncio.sleep(1)
        else:
            return

    logger.error("Failed to register study after 30 attempts")


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, override

import aiofiles.os

from lite_dist2.common import async_read_file, async_write_file
from lite_dist2.curriculum_models.trial import TrialModel
from lite_dist2.trial_repositories.base_trial_repository import BaseTrialRepository
from lite_dist2.trial_repositories.trial_repository_model import TrialRepositoryModel

if TYPE_CHECKING:
    from pathlib import Path

    from lite_dist2.type_definitions import TrialRepositoryType


class NormalTrialRepository(BaseTrialRepository):
    def __init__(self, save_dir: Path) -> None:
        self.save_dir = save_dir

    @override
    @staticmethod
    def get_repository_type() -> TrialRepositoryType:
        return "normal"

    @override
    async def clean_save_dir(self) -> None:
        if self.save_dir.exists():
            for item in self.save_dir.iterdir():
                if item.is_file() or item.is_symlink():
                    await aiofiles.os.remove(item)
                else:
                    warnings.warn(f"Unexpected item in save_dir: {item}", stacklevel=2)
        else:
            if not self.save_dir.parent.exists():
                await aiofiles.os.mkdir(self.save_dir.parent)
            await aiofiles.os.mkdir(self.save_dir)

    @override
    async def save(self, trial: TrialModel) -> None:
        path = self.save_dir / f"{trial.trial_id}.json"
        await async_write_file(path, trial.model_dump_json().encode("utf-8"))

    @override
    async def load(self, trial_id: str) -> TrialModel:
        path = self.save_dir / f"{trial_id}.json"
        content = await async_read_file(path)
        return TrialModel.model_validate_json(content)

    @override
    async def load_all(self) -> list[TrialModel]:
        trials = []
        if not self.save_dir.exists() or not self.save_dir.is_dir():
            raise FileNotFoundError(self.save_dir)

        for json_path in self.save_dir.glob("*.json"):
            content = await async_read_file(json_path)
            trials.append(TrialModel.model_validate_json(content))
        return trials

    @override
    async def delete_save_dir(self) -> None:
        # ディレクトリ内の全ファイルを削除
        if self.save_dir.exists() and self.save_dir.is_dir():
            for item in self.save_dir.iterdir():
                if item.is_file() or item.is_symlink():
                    await aiofiles.os.remove(item)
            # ディレクトリ自体を削除
            await aiofiles.os.rmdir(self.save_dir)

    @override
    def to_model(self) -> TrialRepositoryModel:
        return TrialRepositoryModel(type=self.get_repository_type(), save_dir=self.save_dir)

    @staticmethod
    def from_model(model: TrialRepositoryModel) -> NormalTrialRepository:
        return NormalTrialRepository(model.save_dir)



import pytest

from lite_dist2.expections import LD2UndefinedError
from lite_dist2.value_models.const_param import ConstParam, ConstParamElement


def test_const_param_to_dict_from_dict() -> None:
    original = ConstParam(
        consts=[
            ConstParamElement(type="int", key="int_value", value="0x2"),
            ConstParamElement(type="float", key="float_value", value="0x1.0000000000000p+0"),
            ConstParamElement(type="bool", key="bool_value", value=False),
            ConstParamElement(type="str", key="str_value", value="rk"),
        ],
    )

    d = original.to_dict()
    actual = ConstParam.from_dict(d)
    assert original == actual


def test_const_param_element_from_kv_raise_undefined_type() -> None:
    with pytest.raises(LD2UndefinedError):
        _ = ConstParamElement.from_kv(key="error_value", value=[])  # ty: ignore[invalid-argument-type]



from __future__ import annotations

from typing import TYPE_CHECKING

from lite_dist2.expections import LD2UndefinedError
from lite_dist2.trial_repositories.normal_trial_repository import NormalTrialRepository

if TYPE_CHECKING:
    from lite_dist2.trial_repositories.base_trial_repository import BaseTrialRepository
    from lite_dist2.trial_repositories.trial_repository_model import TrialRepositoryModel


def create_trial_repository(model: TrialRepositoryModel) -> BaseTrialRepository:
    match model.type:
        case "normal":
            return NormalTrialRepository.from_model(model)
        case _:
            raise LD2UndefinedError(model.type)



import json
from pathlib import Path

import pytest

from lite_dist2.curriculum_models.trial import TrialModel, TrialStatus
from lite_dist2.trial_repositories.normal_trial_repository import NormalTrialRepository
from lite_dist2.value_models.aligned_space import ParameterAlignedSpacePortableModel
from lite_dist2.value_models.line_segment import LineSegmentPortableModel
from tests.const import DT


def test_normal_trial_repository_get_repository_type() -> None:
    repo = NormalTrialRepository(Path("test/s01"))
    actual = repo.get_repository_type()
    expected = "normal"
    assert actual == expected


@pytest.mark.asyncio
async def test_normal_trial_repository_clean_save_dir_empty(tmp_path: str) -> None:
    save_dir = Path(tmp_path) / "some_study"

    # 空であることを確認
    assert not save_dir.exists()

    repo = NormalTrialRepository(save_dir)
    await repo.clean_save_dir()
    assert save_dir.exists()
    assert save_dir.is_dir()
    assert not bool(list(save_dir.iterdir()))


@pytest.mark.asyncio
async def test_normal_trial_repository_clean_save_dir_filled(tmp_path: str) -> None:
    save_dir = Path(tmp_path) / "some_study"

    # 空であることを確認
    assert not save_dir.exists()

    # ディレクトリとファイルを作成する
    save_dir.mkdir(parents=False, exist_ok=False)
    dummy_file = save_dir / "dummy.txt"
    dummy_file.touch(exist_ok=False)
    assert bool(list(save_dir.iterdir()))

    repo = NormalTrialRepository(save_dir)
    await repo.clean_save_dir()

    # 初期状態が空じゃなくても空になることを確認
    assert save_dir.exists()
    assert save_dir.is_dir()
    assert not bool(list(save_dir.iterdir()))


@pytest.mark.asyncio
async def test_normal_trial_repository_save(tmp_path: str) -> None:
    save_dir = Path(tmp_path) / "s01"
    trial_model = TrialModel(
        study_id="s01",
        trial_id="t01",
        reserved_timestamp=DT,
        trial_status=TrialStatus.running,
        const_param=None,
        parameter_space=ParameterAlignedSpacePortableModel(
            type="aligned",
            axes=[
                LineSegmentPortableModel(
                    name="x",
                    type="int",
                    size="0x64",
                    step="0x1",
                    start="0x0",
                    ambient_size="0x64",
                    ambient_index="0x0",
                ),
            ],
            check_lower_filling=True,
        ),
        result_type="scalar",
        result_value_type="int",
        worker_node_id="n01",
        worker_node_name="n01",
        registered_timestamp=DT,
    )

    repo = NormalTrialRepository(save_dir)
    await repo.clean_save_dir()

    # save
    await repo.save(trial_model)

    # save されているか確認
    trial_file = save_dir / "t01.json"
    assert trial_file.exists()

    # マニュアル load
    with trial_file.open() as f:
        d = json.load(f)
        loaded_trial_model = TrialModel.model_validate(d)

    # 同じものか確認
    assert trial_model == loaded_trial_model


@pytest.mark.asyncio
async def test_normal_trial_repository_load(tmp_path: str) -> None:
    save_dir = Path(tmp_path) / "s01"
    trial_model = TrialModel(
        study_id="s01",
        trial_id="t01",
        reserved_timestamp=DT,
        trial_status=TrialStatus.running,
        const_param=None,
        parameter_space=ParameterAlignedSpacePortableModel(
            type="aligned",
            axes=[
                LineSegmentPortableModel(
                    name="x",
                    type="int",
                    size="0x64",
                    step="0x1",
                    start="0x0",
                    ambient_size="0x64",
                    ambient_index="0x0",
                ),
            ],
            check_lower_filling=True,
        ),
        result_type="scalar",
        result_value_type="int",
        worker_node_id="n01",
        worker_node_name="n01",
        registered_timestamp=DT,
    )

    repo = NormalTrialRepository(save_dir)
    await repo.clean_save_dir()

    # マニュアル save
    trial_file = save_dir / "t01.json"
    with trial_file.open("w") as f:
        d = trial_model.model_dump(mode="json")
        json.dump(d, f)

    # save されているか確認
    trial_file = save_dir / "t01.json"
    assert trial_file.exists()

    # load
    loaded_trial_model = await repo.load("t01")

    # 同じものか確認
    assert trial_model == loaded_trial_model


@pytest.mark.asyncio
async def test_normal_trial_repository_load_all(tmp_path: str) -> None:
    save_dir = Path(tmp_path) / "s01"
    trial_models = [
        TrialModel(
            study_id="s01",
            trial_id=f"t0{i}",
            reserved_timestamp=DT,
            trial_status=TrialStatus.running,
            const_param=None,
            parameter_space=ParameterAlignedSpacePortableModel(
                type="aligned",
                axes=[
                    LineSegmentPortableModel(
                        name="x",
                        type="int",
                        size="0x64",
                        step="0x1",
                        start="0x0",
                        ambient_size="0x64",
                        ambient_index="0x0",
                    ),
                ],
                check_lower_filling=True,
            ),
            result_type="scalar",
            result_value_type="int",
            worker_node_id="n01",
            worker_node_name="n01",
            registered_timestamp=DT,
        )
        for i in range(1, 4)
    ]

    # マニュアル save
    repo = NormalTrialRepository(save_dir)
    await repo.clean_save_dir()
    for trial in trial_models:
        path = save_dir / f"{trial.trial_id}.json"
        with path.open("w") as f:
            json.dump(trial.model_dump(mode="json"), f)

    loaded_trials = await repo.load_all()
    assert loaded_trials == trial_models


@pytest.mark.asyncio
async def test_normal_trial_repository_delete_save_dir_not_exist(tmp_path: str) -> None:
    save_dir = Path(tmp_path) / "some_study"

    # 空であることを確認
    assert not save_dir.exists()

    repo = NormalTrialRepository(save_dir)
    await repo.delete_save_dir()

    assert not save_dir.exists()


@pytest.mark.asyncio
async def test_normal_trial_repository_delete_save_dir_exist(tmp_path: str) -> None:
    save_dir = Path(tmp_path) / "some_study"

    # 空であることを確認
    assert not save_dir.exists()

    # ディレクトリとファイルを作成する
    save_dir.mkdir(parents=False, exist_ok=False)
    dummy_file = save_dir / "dummy.txt"
    dummy_file.touch(exist_ok=False)
    assert bool(list(save_dir.iterdir()))

    repo = NormalTrialRepository(save_dir)
    await repo.delete_save_dir()

    assert not save_dir.exists()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.common import hex2float, hex2int, int2hex, portablize
from lite_dist2.expections import LD2InvalidSpaceError, LD2ModelTypeError, LD2ParameterError
from lite_dist2.type_definitions import PortableValueType, PrimitiveValueType

if TYPE_CHECKING:
    from collections.abc import Generator


class LineSegment[T: PrimitiveValueType]:
    def __init__(
        self,
        name: str | None,
        type_: Literal["bool", "int", "float"],
        size: int | None,
        start: T,
        step: T,
        ambient_index: int,
        ambient_size: int | None,
        is_dummy: bool = False,
    ) -> None:
        self.name = name
        self.type = type_
        self.size = size
        self.start = start
        self.step = step
        self.ambient_index = ambient_index
        self.ambient_size = ambient_size
        self.is_dummy = is_dummy

        self._dtype = type(self.start)

    def __hash__(self) -> int:
        return hash((self.name, self.type, self.size, self.start, self.step, self.ambient_size, self.ambient_index))

    def grid(self) -> Generator[T]:
        fstart = float(self.start)
        fstep = float(self.step)
        i = 0
        while self.size is None or i < self.size:
            yield self._dtype(fstart + i * fstep)
            i += 1

    def indexed_grid(self) -> Generator[tuple[int, T]]:
        i = 0
        while self.size is None or i < self.size:
            yield i + self.ambient_index, self.start + i * self.step
            i += 1

    def slice(self, start_index: int, size: int) -> LineSegment[T]:
        if self.size is not None and size > self.size - start_index:
            msg = f"{size=}"
            raise LD2ParameterError(msg, "larger than ambient")
        return LineSegment[T](
            name=self.name,
            type_=self.type,
            start=self._dtype(self.start + start_index * self.step),
            size=size,
            step=self.step,
            ambient_index=self.ambient_index + start_index,
            ambient_size=self.ambient_size,
        )

    def get_step(self) -> T:
        return self.step

    def get_start_index(self, *_: object) -> int:
        return self.ambient_index

    def merge(self, other: LineSegment, *_: object) -> LineSegment:
        smaller, larger = (self, other) if self.ambient_index < other.ambient_index else (other, self)
        size = larger.end_index() - smaller.ambient_index + 1
        return LineSegment[T](
            name=self.name,
            type_=self.type,
            size=size,
            ambient_index=smaller.ambient_index,
            ambient_size=self.ambient_size,
            start=smaller.start,
            step=self.step,
        )

    def can_merge(self, other: LineSegment, *_: object) -> bool:
        if self.ambient_index < other.ambient_index:
            smaller = self
            larger = other
        else:
            smaller = other
            larger = self

        return smaller.end_index() + 1 >= larger.ambient_index

    def end_index(self) -> int:
        if self.size is None:
            msg = "Cannot get end index because this axis is infinite."
            raise LD2InvalidSpaceError(msg)
        return self.ambient_index + self.size - 1

    def is_universal(self) -> bool:
        if self.ambient_size is None:
            return False
        return self.size == self.ambient_size

    def derived_by_same_ambient_space_with(self, other: LineSegment) -> bool:
        return (
            (self.name == other.name)
            and (self.type == other.type)
            and (self.get_step() == other.get_step())
            and (self.ambient_size == other.ambient_size)
        )

    def to_dummy(self) -> DummyLineSegment:
        return DummyLineSegment(
            name=self.name,
            type_=self.type,
            step=self.step,
            ambient_size=self.ambient_size,
        )


class DummyLineSegment:
    def __init__(
        self,
        type_: Literal["bool", "int", "float"],
        name: str | None,
        step: PrimitiveValueType,
        ambient_size: int | None,
    ) -> None:
        self.type = type_
        self.name = name
        self.step = step
        self.ambient_size = ambient_size


class LineSegmentPortableModel(BaseModel):
    name: str | None = None
    type: Literal["bool", "int", "float"]
    size: str | None
    step: PortableValueType
    start: PortableValueType
    ambient_index: str
    ambient_size: str | None = None
    is_dummy: bool = False

    def to_line_segment(self) -> LineSegment:
        if self.is_dummy:
            return LineSegment(
                name=self.name or "dummy",
                type_=self.type,
                size=1,
                step=self.step if isinstance(self.step, bool) else hex2int(self.step),
                start=0,
                ambient_index=0,
                ambient_size=None if self.ambient_size is None else hex2int(self.ambient_size),
                is_dummy=True,
            )
        match self.type:
            case "bool":
                if self.size is None or self.ambient_size is None:
                    p = "size or ambient_size"
                    e = "size and ambient_size are required for bool type"
                    raise LD2ParameterError(p, e)
                return LineSegment(
                    name=self.name,
                    type_="bool",
                    size=hex2int(self.size),
                    step=int(self.step),
                    start=bool(self.start),
                    ambient_index=hex2int(self.ambient_index),
                    ambient_size=None if self.ambient_size is None else hex2int(self.ambient_size),
                    is_dummy=False,
                )
            case "int":
                return LineSegment(
                    name=self.name,
                    type_=self.type,
                    size=None if self.size is None else hex2int(self.size),
                    start=hex2int(str(self.start)),
                    step=hex2int(str(self.step)),
                    ambient_index=hex2int(self.ambient_index),
                    ambient_size=None if self.ambient_size is None else hex2int(self.ambient_size),
                    is_dummy=False,
                )
            case "float":
                return LineSegment(
                    name=self.name,
                    type_=self.type,
                    size=None if self.size is None else hex2int(self.size),
                    start=hex2float(str(self.start)),
                    step=hex2float(str(self.step)),
                    ambient_index=hex2int(self.ambient_index),
                    ambient_size=None if self.ambient_size is None else hex2int(self.ambient_size),
                    is_dummy=False,
                )
            case _:
                msg = f"Invalid line segment type: {self.type}"
                raise LD2InvalidSpaceError(msg)

    @staticmethod
    def from_line_segment(line_segment: LineSegment | DummyLineSegment) -> LineSegmentPortableModel:
        match line_segment:
            case DummyLineSegment():
                return LineSegmentPortableModel(
                    name=line_segment.name,
                    type=line_segment.type,
                    size=None,
                    start=portablize(line_segment.type, 0),
                    step=portablize(line_segment.type, line_segment.step),
                    ambient_index=int2hex(0),
                    ambient_size=None if line_segment.ambient_size is None else int2hex(line_segment.ambient_size),
                    is_dummy=True,
                )
            case LineSegment():
                return LineSegmentPortableModel(
                    name=line_segment.name,
                    type=line_segment.type,
                    size=None if line_segment.size is None else int2hex(line_segment.size),
                    start=portablize(line_segment.type, line_segment.start),
                    step=portablize(line_segment.type, line_segment.step),
                    ambient_index=int2hex(line_segment.ambient_index),
                    ambient_size=None if line_segment.ambient_size is None else int2hex(line_segment.ambient_size),
                    is_dummy=line_segment.is_dummy,
                )
            case _:
                raise LD2ModelTypeError(line_segment)



import pytest

from lite_dist2.curriculum_models.mapping import Mapping
from lite_dist2.type_definitions import PortableValueType
from lite_dist2.value_models.point import ScalarValue, VectorValue


@pytest.mark.parametrize(
    ("mapping", "expected"),
    [
        pytest.param(
            Mapping(
                params=(
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                ),
                result=ScalarValue(type="scalar", value_type="int", value="0x2"),
            ),
            ("0x0", "0x1", "0x2"),
            id="scalar",
        ),
        pytest.param(
            Mapping(
                params=(
                    ScalarValue(type="scalar", value_type="bool", value=True, name="x"),
                    ScalarValue(type="scalar", value_type="float", value="0x1.0000000000000p+0", name="y"),
                ),
                result=VectorValue(type="vector", value_type="int", values=["0x2", "0x3"]),
            ),
            (True, "0x1.0000000000000p+0", "0x2", "0x3"),
            id="scalar",
        ),
    ],
)
def test_mapping_to_tuple(mapping: Mapping, expected: tuple[PortableValueType]) -> None:
    actual = mapping.to_tuple()
    assert actual == expected



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import httpx

from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import Trial
from lite_dist2.expections import LD2TableNodeServerError
from lite_dist2.table_node_api.table_param import StudyRegisterParam, TrialRegisterParam, TrialReserveParam
from lite_dist2.table_node_api.table_response import (
    OkResponse,
    StudyRegisteredResponse,
    StudyResponse,
    TrialReserveResponse,
)

if TYPE_CHECKING:
    from typing import Any, ClassVar

    from pydantic import BaseModel

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class TableNodeClient:
    INSTANT_API_TIMEOUT_SECONDS = 10
    HEADERS: ClassVar[dict[str, str]] = {"Content-Type": "application/json; charset=utf-8"}

    def __init__(self, ip: str, port: int | str) -> None:
        self.domain = f"http://{ip}:{port}"

    async def ping(self) -> bool:
        try:
            _ = await self._get("/ping", timeout_seconds=self.INSTANT_API_TIMEOUT_SECONDS)
        except LD2TableNodeServerError:
            return False
        return True

    async def register_study(self, param: StudyRegisterParam) -> StudyRegisteredResponse:
        status_code, d = await self._post("/study/register", self.INSTANT_API_TIMEOUT_SECONDS, param)
        if status_code != 200:
            msg = f"Failed to register study. status_code={status_code}, response={d}"
            raise LD2TableNodeServerError(msg)

        resp = StudyRegisteredResponse.model_validate(d)
        logger.info("Registered study: %s", resp.study_id)
        return resp

    async def reserve_trial(
        self,
        worker_id: str,
        worker_name: str | None,
        max_size: int,
        retaining_capacity: set[str],
        timeout_seconds: int,
    ) -> Trial | None:
        param = TrialReserveParam(
            retaining_capacity=retaining_capacity,
            max_size=max_size,
            worker_node_name=worker_name,
            worker_node_id=worker_id,
        )
        status_code, d = await self._post("/trial/reserve", timeout_seconds, param)

        resp = TrialReserveResponse.model_validate(d)
        if status_code == 202 or resp.trial is None:
            logger.info("Cannot reserve trial")
            return None

        trial = Trial.from_model(resp.trial)
        logger.info("Reserved trial (size=%d)", trial.parameter_space.total)
        return trial

    async def register_trial(self, trial: Trial, timeout_seconds: int) -> None:
        param = TrialRegisterParam(trial=trial.to_model())
        status_code, _ = await self._post("/trial/register", timeout_seconds, param)
        if status_code == 409:
            logger.warning("Failed to register trial. This trial might be timed out or study might be cancelled.")
        elif status_code != 200:
            logger.warning("Failed to register trial.")

    async def study(self, study_id: str | None = None, name: str | None = None) -> StudyResponse | None:
        _, resp = await self._get("/study", self.INSTANT_API_TIMEOUT_SECONDS, {"study_id": study_id, "name": name})
        study_response = StudyResponse.model_validate(resp)
        if study_response.status != StudyStatus.done:
            detail_info = f"{study_id=}" if study_id is not None else f"{name=}"
            logger.info("Study(%s) is %s", detail_info, str(study_response))
        return study_response

    async def save(self) -> OkResponse:
        _, resp = await self._get("/save", self.INSTANT_API_TIMEOUT_SECONDS)
        return OkResponse.model_validate(resp)

    async def _get(
        self,
        path: str,
        timeout_seconds: int,
        query: dict[str, str | None] | None = None,
    ) -> tuple[int, dict[str, Any]]:
        url = f"{self.domain}{path}"
        _query = None if query is None else {k: v for k, v in query.items() if v is not None}
        async with httpx.AsyncClient() as client:
            response = await client.get(
                url,
                headers=self.HEADERS,
                params=_query,
                timeout=timeout_seconds,
            )
        return response.status_code, response.json()

    async def _post(self, path: str, timeout_seconds: int, body: BaseModel) -> tuple[int, dict[str, Any]]:
        url = f"{self.domain}{path}"
        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers=self.HEADERS,
                json=body.model_dump(mode="json"),
                timeout=timeout_seconds,
            )
        response.raise_for_status()
        return response.status_code, response.json()



from __future__ import annotations

from typing import TYPE_CHECKING, override

from lite_dist2.curriculum_models.mapping import MappingsStorage
from lite_dist2.expections import LD2NotDoneError
from lite_dist2.study_strategies import BaseStudyStrategy, StudyStrategyModel

if TYPE_CHECKING:
    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.study_strategies.base_study_strategy import StudyStrategyParam
    from lite_dist2.trial_repositories.base_trial_repository import BaseTrialRepository
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace


class AllCalculationStudyStrategy(BaseStudyStrategy):
    def __init__(self, study_strategy_param: StudyStrategyParam | None) -> None:
        self.study_strategy_param = study_strategy_param

    @override
    async def is_done(
        self,
        trial_table: TrialTable,
        parameter_space: ParameterAlignedSpace,
        trial_repository: BaseTrialRepository,
    ) -> bool:
        return trial_table.count_grid() == parameter_space.total

    @override
    async def extract_mappings(self, trial_repository: BaseTrialRepository) -> MappingsStorage:
        trials = await trial_repository.load_all()
        if not trials:
            raise LD2NotDoneError

        mappings = trials[0].results
        if mappings is None:
            raise LD2NotDoneError
        first = mappings[0]

        params = tuple(param.to_dummy() for param in first.params)
        result = first.result.to_dummy()

        values = []
        for trial in trials:
            if trial.results is None:
                raise LD2NotDoneError
            values.extend(mapping.to_tuple() for mapping in trial.results)
        return MappingsStorage(params_info=params, result_info=result, values=values)

    @override
    def to_model(self) -> StudyStrategyModel:
        return StudyStrategyModel(
            type="all_calculation",
            study_strategy_param=self.study_strategy_param,
        )



import asyncio
import signal
from concurrent.futures import ProcessPoolExecutor

from lite_dist2.config import WorkerConfig
from lite_dist2.type_definitions import RawParamType, RawResultType
from lite_dist2.worker_node.trial_runner import SemiAutoMPTrialRunner
from lite_dist2.worker_node.worker import Worker


def worker_init() -> None:
    # Ignore KeyboardInterrupt on worker processes
    signal.signal(signal.SIGINT, signal.SIG_IGN)


class Mandelbrot(SemiAutoMPTrialRunner):
    def func(self, parameters: RawParamType, *_: object, **kwargs: object) -> RawResultType:
        abs_threshold = self.get_typed("abs_threshold", float, kwargs)
        max_iter = self.get_typed("max_iter", int, kwargs)
        x = float(parameters[0])
        y = float(parameters[1])
        c = complex(x, y)
        z = complex(0, 0)
        iter_count = 0
        while abs(z) <= abs_threshold and iter_count < max_iter:
            z = z**2 + c
            iter_count += 1
        return iter_count


async def main() -> None:
    worker_config = WorkerConfig(
        name="w_01",
        process_num=2,
        max_size=10,
        wait_seconds_on_no_trial=5,
        table_node_request_timeout_seconds=60,
    )
    with ProcessPoolExecutor(max_workers=2, initializer=worker_init) as pool:
        worker = Worker(
            trial_runner=Mandelbrot(),
            ip="table",  # Docker Compose でのサービス名
            port=8000,
            pool=pool,
            config=worker_config,
        )
        try:
            await worker.start_async(stop_at_no_trial=True)
        except KeyboardInterrupt:
            pool.shutdown(wait=False, cancel_futures=True)


if __name__ == "__main__":
    asyncio.run(main())



import os
from pathlib import Path

from lite_dist2.config import TableConfig, TableConfigProvider
from lite_dist2.table_node_api.start_table_api import start


def set_table_config() -> None:
    curriculum_dir = os.getenv("CURRICULUM_DIR", "/app")
    curriculum_path = Path(curriculum_dir) / "curriculum.json"

    config = TableConfig(
        port=8000,
        curriculum_path=curriculum_path,
        curriculum_save_interval_seconds=10,
    )
    TableConfigProvider.set(config)


if __name__ == "__main__":
    set_table_config()
    start()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.common import numerize, portablize
from lite_dist2.expections import LD2ModelTypeError
from lite_dist2.type_definitions import PortableValueType

if TYPE_CHECKING:
    from collections.abc import Iterable

    from lite_dist2.type_definitions import PrimitiveValueType


def _get_default_value(value_type: Literal["bool", "int", "float"]) -> PortableValueType:
    match value_type:
        case "bool":
            return False
        case "int":
            return "0x0"
        case "float":
            return "0x0.0p+0"
        case _:
            raise LD2ModelTypeError(value_type)


class ScalarValue(BaseModel):
    type: Literal["scalar"]
    value_type: Literal["bool", "int", "float"]
    value: PortableValueType
    name: str | None = None

    def numerize(self) -> PrimitiveValueType:
        return numerize(self.value_type, self.value)

    def equal_to(self, other: object) -> bool:
        if not isinstance(other, ScalarValue):
            return False
        return self.value_type == other.value_type and self.value == other.value

    def get_value_size(self) -> int:
        return 1

    def get_value_types(self) -> tuple[Literal["bool", "int", "float"], ...]:
        return (self.value_type,)

    def get_value_list(self) -> list[PortableValueType]:
        return [self.value]

    def to_dummy(self) -> ScalarValue:
        return ScalarValue(
            type="scalar",
            value_type=self.value_type,
            value=_get_default_value(self.value_type),
            name=self.name,
        )

    @staticmethod
    def create_from_numeric(
        raw_result_value: PrimitiveValueType,
        value_type: Literal["bool", "int", "float"],
        name: str | None = None,
    ) -> ScalarValue:
        val = portablize(value_type, raw_result_value)
        return ScalarValue(type="scalar", value_type=value_type, value=val, name=name)


class VectorValue(BaseModel):
    type: Literal["vector"]
    value_type: Literal["bool", "int", "float"]
    values: list[PortableValueType]
    name: str | None = None

    def numerize(self) -> list[PrimitiveValueType]:
        return [numerize(self.value_type, v) for v in self.values]

    def equal_to(self, other: object) -> bool:
        if not isinstance(other, VectorValue):
            return False
        return self.value_type == other.value_type and self.values == other.values

    def get_value_size(self) -> int:
        return len(self.values)

    def get_value_types(self) -> tuple[Literal["bool", "int", "float"], ...]:
        return (self.value_type,) * self.get_value_size()

    def get_value_list(self) -> list[PortableValueType]:
        return self.values

    def to_dummy(self) -> VectorValue:
        return VectorValue(
            type="vector",
            value_type=self.value_type,
            values=[_get_default_value(self.value_type) for _ in self.values],
            name=self.name,
        )

    @staticmethod
    def create_from_numeric(
        raw_result_value: Iterable[PrimitiveValueType],
        value_type: Literal["bool", "int", "float"],
        name: str | None = None,
    ) -> VectorValue:
        val = [portablize(value_type, rv) for rv in raw_result_value]
        return VectorValue(type="vector", value_type=value_type, values=val, name=name)


type ParamType = tuple[ScalarValue, ...]
type ResultType = ScalarValue | VectorValue



from __future__ import annotations

from datetime import datetime
from enum import StrEnum, auto
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.common import publish_timestamp
from lite_dist2.curriculum_models.mapping import Mapping
from lite_dist2.expections import LD2InvalidSpaceError, LD2ModelTypeError, LD2NotDoneError, LD2UndefinedError
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpacePortableModel
from lite_dist2.value_models.const_param import ConstParam
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace, ParameterJaggedSpacePortableModel
from lite_dist2.value_models.point import ResultType, ScalarValue, VectorValue
from lite_dist2.value_models.space_model import SpacePortableModelType

if TYPE_CHECKING:
    from lite_dist2.type_definitions import RawParamType, RawResultType
    from lite_dist2.value_models.base_space import FlattenSegment
    from lite_dist2.value_models.space_type import ParameterSpaceType


class TrialStatus(StrEnum):
    running = auto()
    done = auto()


class TrialDoneRecord(BaseModel):
    trial_id: str
    reserved_timestamp: datetime
    worker_node_name: str | None
    worker_node_id: str
    registered_timestamp: datetime
    grid_size: int

    def calc_duration_sec(self) -> float:
        dt = self.registered_timestamp - self.reserved_timestamp
        return dt.total_seconds()

    def calc_grid_per_sec(self) -> float:
        return self.grid_size / self.calc_duration_sec()


class TrialModel(BaseModel):
    study_id: str
    trial_id: str
    reserved_timestamp: datetime
    trial_status: TrialStatus
    const_param: ConstParam | None
    parameter_space: SpacePortableModelType
    result_type: Literal["scalar", "vector"]
    result_value_type: Literal["bool", "int", "float"]
    worker_node_name: str | None
    worker_node_id: str
    results: list[Mapping] | None = None
    registered_timestamp: datetime | None = None


class Trial:
    def __init__(
        self,
        study_id: str,
        trial_id: str,
        reserved_timestamp: datetime,
        trial_status: TrialStatus,
        const_param: ConstParam | None,
        parameter_space: ParameterSpaceType,
        result_type: Literal["scalar", "vector"],
        result_value_type: Literal["bool", "int", "float"],
        worker_node_name: str | None,
        worker_node_id: str,
        results: list[Mapping] | None = None,
        registered_timestamp: datetime | None = None,
    ) -> None:
        self.study_id = study_id
        self.trial_id = trial_id
        self.reserved_timestamp = reserved_timestamp
        self.trial_status = trial_status
        self.const_param = const_param
        self.parameter_space = parameter_space
        self.result_type = result_type
        self.result_value_type = result_value_type
        self.worker_node_name = worker_node_name
        self.worker_node_id = worker_node_id
        self.result = results
        self.registered_timestamp = registered_timestamp

    def convert_mappings_from(self, raw_mappings: list[tuple[RawParamType, RawResultType]]) -> list[Mapping]:
        mappings = []
        for raw_param, raw_res in raw_mappings:
            param = self.parameter_space.value_tuple_to_param_type(raw_param)
            result = self._create_result_value(raw_res)
            mappings.append(Mapping(params=param, result=result))
        return mappings

    def set_result(self, mappings: list[Mapping]) -> None:
        self.result = mappings

    def set_registered_timestamp(self) -> None:
        self.registered_timestamp = publish_timestamp()

    def get_running_segments(self) -> list[FlattenSegment]:
        if self.trial_status == TrialStatus.running:
            return self.parameter_space.get_flatten_ambient_start_and_size_list()
        return []

    def _create_result_value(self, raw_result: RawResultType) -> ResultType:
        if self.result_type == "scalar" and isinstance(raw_result, (bool, int, float)):
            return ScalarValue.create_from_numeric(raw_result, self.result_value_type)
        if self.result_type == "vector" and isinstance(raw_result, list):
            return VectorValue.create_from_numeric(raw_result, self.result_value_type)
        raise LD2ModelTypeError(self.result_type)

    def measure_seconds_from_registered(self, now: datetime) -> int:
        delta = now - self.reserved_timestamp
        return int(delta.total_seconds())

    def find_target_value(self, target_value: ResultType) -> Mapping | None:
        # find_exact 用
        if not self.result:
            return None
        if self.trial_status != TrialStatus.done:
            return None

        for mapping in self.result:
            if mapping.result.equal_to(target_value):
                return mapping
        return None

    def to_done_record(self) -> TrialDoneRecord:
        if self.trial_status != TrialStatus.done or self.registered_timestamp is None:
            raise LD2NotDoneError

        grid_size = self.parameter_space.total
        if not isinstance(grid_size, int):
            msg = "Parameter space in trial must be finite space."
            raise LD2InvalidSpaceError(msg)
        return TrialDoneRecord(
            trial_id=self.trial_id,
            reserved_timestamp=self.reserved_timestamp,
            worker_node_name=self.worker_node_name,
            worker_node_id=self.worker_node_id,
            registered_timestamp=self.registered_timestamp,
            grid_size=grid_size,
        )

    def done_in_after(self, cutoff_datetime: datetime) -> bool:
        if self.trial_status != TrialStatus.done or self.registered_timestamp is None:
            return False
        return cutoff_datetime < self.registered_timestamp

    def to_model(self) -> TrialModel:
        return TrialModel(
            study_id=self.study_id,
            trial_id=self.trial_id,
            reserved_timestamp=self.reserved_timestamp,
            trial_status=self.trial_status,
            const_param=self.const_param,
            parameter_space=self.parameter_space.to_model(),
            result_type=self.result_type,
            result_value_type=self.result_value_type,
            worker_node_name=self.worker_node_name,
            worker_node_id=self.worker_node_id,
            results=self.result,
            registered_timestamp=self.registered_timestamp,
        )

    @staticmethod
    def from_model(model: TrialModel) -> Trial:
        match model.parameter_space:
            case ParameterAlignedSpacePortableModel():
                parameter_space = ParameterAlignedSpace.from_model(model.parameter_space)
            case ParameterJaggedSpacePortableModel():
                parameter_space = ParameterJaggedSpace.from_model(model.parameter_space)
            case _:
                raise LD2UndefinedError(model.parameter_space.type)
        return Trial(
            study_id=model.study_id,
            trial_id=model.trial_id,
            reserved_timestamp=model.reserved_timestamp,
            trial_status=model.trial_status,
            const_param=model.const_param,
            parameter_space=parameter_space,
            result_type=model.result_type,
            result_value_type=model.result_value_type,
            worker_node_name=model.worker_node_name,
            worker_node_id=model.worker_node_id,
            results=model.results,
            registered_timestamp=model.registered_timestamp,
        )



from collections import defaultdict
from datetime import datetime, timedelta
from typing import Literal

from pydantic import BaseModel

from lite_dist2.curriculum_models.trial import TrialDoneRecord


class WorkerEfficiency(BaseModel):
    worker_id: str
    worker_name: str | None
    grid_velocity: float


class StudyProgressSummary(BaseModel):
    study_id: str
    study_name: str | None
    total_grid: int | Literal["infinite"]
    done_grid: int
    grid_velocity: float
    eta: datetime | Literal["unpredictable"]
    worker_efficiencies: list[WorkerEfficiency]


class ReportMaterial(BaseModel):
    study_id: str
    study_name: str | None
    records: list[TrialDoneRecord]
    total_grid: int | None
    done_grid: int


def report_study_progress(
    now: datetime,
    cutoff_sec: int,
    report_material: ReportMaterial,
) -> StudyProgressSummary:
    records_worker_map = defaultdict(list)
    for rec in report_material.records:
        records_worker_map[rec.worker_node_id].append(rec)

    records_worker_map = dict(records_worker_map)
    worker_efficiencies = []
    for wni, worker_records in records_worker_map.items():
        if not worker_records:
            continue
        worker_efficiencies.append(
            WorkerEfficiency(
                worker_id=wni,
                worker_name=worker_records[0].worker_node_name,
                grid_velocity=sum(rec.grid_size for rec in worker_records) / cutoff_sec,
            ),
        )
    worker_efficiencies = sorted(worker_efficiencies, key=lambda e: e.grid_velocity, reverse=True)
    total_grid_velocity = sum(e.grid_velocity for e in worker_efficiencies)

    if report_material.total_grid is None:
        return StudyProgressSummary(
            study_id=report_material.study_id,
            study_name=report_material.study_name,
            total_grid="infinite",
            done_grid=report_material.done_grid,
            grid_velocity=total_grid_velocity,
            eta="unpredictable",
            worker_efficiencies=worker_efficiencies,
        )
    if total_grid_velocity == 0:
        return StudyProgressSummary(
            study_id=report_material.study_id,
            study_name=report_material.study_name,
            total_grid=report_material.total_grid,
            done_grid=report_material.done_grid,
            grid_velocity=0,
            eta="unpredictable",
            worker_efficiencies=worker_efficiencies,
        )

    eta_sec = (report_material.total_grid - report_material.done_grid) / total_grid_velocity
    return StudyProgressSummary(
        study_id=report_material.study_id,
        study_name=report_material.study_name,
        total_grid=report_material.total_grid,
        done_grid=report_material.done_grid,
        grid_velocity=total_grid_velocity,
        eta=now + timedelta(seconds=eta_sec),
        worker_efficiencies=worker_efficiencies,
    )



from __future__ import annotations

from pydantic import BaseModel, Field

from lite_dist2.curriculum_models.study_portables import StudyRegistry
from lite_dist2.curriculum_models.trial import TrialModel


class BaseParam(BaseModel):
    pass


class StudyRegisterParam(BaseParam):
    study: StudyRegistry = Field(description="`Study` to register")


class TrialReserveParam(BaseParam):
    retaining_capacity: set[str] = Field(description="List of capabilities that the worker node has.")
    max_size: int = Field(description="The maximum size of parameter space reserving.")
    worker_node_name: str | None = Field(default=None, description="Name of the worker node. ")
    worker_node_id: str = Field(description="ID of the worker node")


class TrialRegisterParam(BaseModel):
    trial: TrialModel = Field(description="Registering trial to the table node.")



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, override

import pytest

from lite_dist2.curriculum_models.mapping import Mapping, MappingsStorage
from lite_dist2.curriculum_models.trial import TrialModel, TrialStatus
from lite_dist2.curriculum_models.trial_table import TrialTable
from lite_dist2.expections import LD2NotDoneError
from lite_dist2.study_strategies.all_calculation_study_strategy import AllCalculationStudyStrategy
from lite_dist2.trial_repositories.base_trial_repository import BaseTrialRepository
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpacePortableModel
from lite_dist2.value_models.line_segment import LineSegment, LineSegmentPortableModel
from lite_dist2.value_models.point import ScalarValue, VectorValue
from tests.const import DT

if TYPE_CHECKING:
    from lite_dist2.trial_repositories.trial_repository_model import TrialRepositoryModel
    from lite_dist2.type_definitions import TrialRepositoryType

_DUMMY_PARAMETER_SPACE_MODEL = ParameterAlignedSpacePortableModel(
    type="aligned",
    axes=[
        LineSegmentPortableModel(
            name="x",
            type="int",
            size="0x2",
            start="0x0",
            ambient_size="0x2",
            ambient_index="0x0",
            step="0x1",
        ),
        LineSegmentPortableModel(
            name="y",
            type="int",
            size="0x2",
            start="0x0",
            ambient_size="0x2",
            ambient_index="0x0",
            step="0x1",
        ),
    ],
    check_lower_filling=True,
)


class MockTrialRepository(BaseTrialRepository):
    def __init__(self) -> None:
        self.save_dir = Path("test/s01")

    @override
    @staticmethod
    def get_repository_type() -> TrialRepositoryType:
        return "normal"

    @override
    async def clean_save_dir(self) -> None:
        pass

    @override
    async def save(self, trial: TrialModel) -> None:
        pass

    @override
    async def load(self, trial_id: str) -> TrialModel:
        raise NotImplementedError

    @override
    async def load_all(self) -> list[TrialModel]:
        raise NotImplementedError

    @override
    async def delete_save_dir(self) -> None:
        pass

    @override
    def to_model(self) -> TrialRepositoryModel:
        raise NotImplementedError


@pytest.fixture
def done_grid_fixture(monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest) -> None:
    return_value = request.param

    def fake_count_grid(self) -> int:  # noqa: ANN001
        return return_value

    monkeypatch.setattr(TrialTable, "count_grid", fake_count_grid)


@pytest.fixture
def all_grid_fixture(monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest) -> None:
    return_value = request.param

    monkeypatch.setattr(
        ParameterAlignedSpace,
        "total",
        property(lambda _: return_value),
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("done_grid_fixture", "all_grid_fixture", "expected"),
    [
        pytest.param(10, 10, True, id="Done"),
        pytest.param(9, 10, False, id="Yet"),
    ],
    indirect=["done_grid_fixture", "all_grid_fixture"],
)
async def test_all_calculation_study_strategy_is_done(
    done_grid_fixture: int,
    all_grid_fixture: int,
    expected: bool,
) -> None:
    strategy = AllCalculationStudyStrategy(study_strategy_param=None)
    trial_table = TrialTable(trials=[], aggregated_parameter_space=None)
    parameter_space = ParameterAlignedSpace(
        axes=[
            LineSegment(name="x", type_="int", size=2, step=1, start=0, ambient_size=2, ambient_index=0),
            LineSegment(name="y", type_="int", size=2, step=1, start=0, ambient_size=2, ambient_index=0),
        ],
        check_lower_filling=True,
    )
    actual = await strategy.is_done(trial_table, parameter_space, MockTrialRepository())
    assert actual == expected


@pytest.fixture
def trial_repository_fixture(monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest) -> None:
    return_value = request.param

    async def fake_load_all(self) -> int:  # noqa: ANN001
        return return_value

    monkeypatch.setattr(MockTrialRepository, "load_all", fake_load_all)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("trial_repository_fixture", "expected"),
    [
        pytest.param(
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x67"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            MappingsStorage(
                params_info=(
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                ),
                result_info=ScalarValue(type="scalar", value_type="int", value="0x0"),
                values=[
                    ("0x1", "0x1", "0x67"),
                ],
            ),
            id="Single trial, single map",
        ),
        pytest.param(
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x67"),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x2", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x68"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
                TrialModel(
                    trial_id="t02",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x3", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x3", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x69"),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x4", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x4", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x6a"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            MappingsStorage(
                params_info=(
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                ),
                result_info=ScalarValue(type="scalar", value_type="int", value="0x0"),
                values=[
                    ("0x1", "0x1", "0x67"),
                    ("0x2", "0x2", "0x68"),
                    ("0x3", "0x3", "0x69"),
                    ("0x4", "0x4", "0x6a"),
                ],
            ),
            id="Multi trial, multi map, scalar",
        ),
        pytest.param(
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                            ),
                            result=VectorValue(type="vector", value_type="int", values=["0x67", "0x67"]),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x2", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                            ),
                            result=VectorValue(type="vector", value_type="int", values=["0x68", "0x68"]),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
                TrialModel(
                    trial_id="t02",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x3", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x3", name="y"),
                            ),
                            result=VectorValue(type="vector", value_type="int", values=["0x69", "0x69"]),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x4", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x4", name="y"),
                            ),
                            result=VectorValue(type="vector", value_type="int", values=["0x6a", "0x6a"]),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            MappingsStorage(
                params_info=(
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                ),
                result_info=VectorValue(type="vector", value_type="int", values=["0x0", "0x0"]),
                values=[
                    ("0x1", "0x1", "0x67", "0x67"),
                    ("0x2", "0x2", "0x68", "0x68"),
                    ("0x3", "0x3", "0x69", "0x69"),
                    ("0x4", "0x4", "0x6a", "0x6a"),
                ],
            ),
            id="Multi trial, multi map, vector",
        ),
    ],
    indirect=["trial_repository_fixture"],
)
async def test_all_calculation_study_strategy_extract_mapping(
    trial_repository_fixture: list[TrialModel],
    expected: MappingsStorage,
) -> None:
    repo = MockTrialRepository()
    strategy = AllCalculationStudyStrategy(None)
    actual = await strategy.extract_mappings(repo)
    assert actual == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "trial_repository_fixture",
    [
        pytest.param(
            [],
            id="Empty",
        ),
        pytest.param(
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=None,
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
                TrialModel(
                    trial_id="t02",
                    trial_status=TrialStatus.done,
                    results=None,
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            id="None first",
        ),
        pytest.param(
            [
                TrialModel(
                    trial_id="t01",
                    trial_status=TrialStatus.done,
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x67"),
                        ),
                        Mapping(
                            params=(
                                ScalarValue(type="scalar", value_type="int", value="0x2", name="x"),
                                ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                            ),
                            result=ScalarValue(type="scalar", value_type="int", value="0x68"),
                        ),
                    ],
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
                TrialModel(
                    trial_id="t02",
                    trial_status=TrialStatus.done,
                    results=None,
                    study_id="s01",
                    reserved_timestamp=DT,
                    const_param=None,
                    parameter_space=_DUMMY_PARAMETER_SPACE_MODEL,
                    result_type="scalar",
                    result_value_type="int",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            id="None",
        ),
    ],
    indirect=["trial_repository_fixture"],
)
async def test_all_calculation_study_strategy_extract_mapping_raise(
    trial_repository_fixture: list[TrialModel],
) -> None:
    repo = MockTrialRepository()
    strategy = AllCalculationStudyStrategy(None)
    with pytest.raises(LD2NotDoneError):
        _ = await strategy.extract_mappings(repo)



import abc

from lite_dist2.curriculum_models.trial import TrialModel
from lite_dist2.trial_repositories.trial_repository_model import TrialRepositoryModel
from lite_dist2.type_definitions import TrialRepositoryType


class BaseTrialRepository(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def get_repository_type() -> TrialRepositoryType:
        pass

    @abc.abstractmethod
    async def clean_save_dir(self) -> None:
        pass

    @abc.abstractmethod
    async def save(self, trial: TrialModel) -> None:
        pass

    @abc.abstractmethod
    async def load(self, trial_id: str) -> TrialModel:
        pass

    @abc.abstractmethod
    async def load_all(self) -> list[TrialModel]:
        pass

    @abc.abstractmethod
    async def delete_save_dir(self) -> None:
        pass

    @abc.abstractmethod
    def to_model(self) -> TrialRepositoryModel:
        pass



from datetime import datetime

import pytest

from lite_dist2.curriculum_models.progress_summary import (
    ReportMaterial,
    StudyProgressSummary,
    WorkerEfficiency,
    report_study_progress,
)
from lite_dist2.curriculum_models.trial import TrialDoneRecord
from tests.const import JST

NOW = datetime(2025, 6, 28, 19, 45, 0, tzinfo=JST)


@pytest.mark.parametrize(
    ("report_material", "expected"),
    [
        pytest.param(
            ReportMaterial(
                study_id="s01",
                study_name="s01",
                records=[
                    TrialDoneRecord(
                        trial_id="t01",
                        reserved_timestamp=datetime(2025, 6, 28, 19, 40, 0, tzinfo=JST),
                        registered_timestamp=datetime(2025, 6, 28, 19, 42, 0, tzinfo=JST),
                        worker_node_id="w01",
                        worker_node_name="w01",
                        grid_size=100,
                    ),
                    TrialDoneRecord(
                        trial_id="t02",
                        reserved_timestamp=datetime(2025, 6, 28, 19, 42, 0, tzinfo=JST),
                        registered_timestamp=datetime(2025, 6, 28, 19, 44, 0, tzinfo=JST),
                        worker_node_id="w01",
                        worker_node_name="w01",
                        grid_size=100,
                    ),
                ],
                total_grid=None,
                done_grid=400,
            ),
            StudyProgressSummary(
                study_id="s01",
                study_name="s01",
                total_grid="infinite",
                done_grid=400,
                grid_velocity=0.4,
                eta="unpredictable",
                worker_efficiencies=[
                    WorkerEfficiency(worker_id="w01", worker_name="w01", grid_velocity=0.4),
                ],
            ),
            id="Infinite space",
        ),
        pytest.param(
            ReportMaterial(
                study_id="s01",
                study_name="s01",
                records=[],
                total_grid=1000,
                done_grid=400,
            ),
            StudyProgressSummary(
                study_id="s01",
                study_name="s01",
                total_grid=1000,
                done_grid=400,
                grid_velocity=0,
                eta="unpredictable",
                worker_efficiencies=[],
            ),
            id="Stopped study",
        ),
        pytest.param(
            ReportMaterial(
                study_id="s01",
                study_name="s01",
                records=[
                    TrialDoneRecord(
                        trial_id="t01",
                        reserved_timestamp=datetime(2025, 6, 28, 19, 40, 0, tzinfo=JST),
                        registered_timestamp=datetime(2025, 6, 28, 19, 42, 0, tzinfo=JST),
                        worker_node_id="w01",
                        worker_node_name="w01",
                        grid_size=100,
                    ),
                    TrialDoneRecord(
                        trial_id="t03",
                        reserved_timestamp=datetime(2025, 6, 28, 19, 41, 0, tzinfo=JST),
                        registered_timestamp=datetime(2025, 6, 28, 19, 43, 0, tzinfo=JST),
                        worker_node_id="w02",
                        worker_node_name="w02",
                        grid_size=50,
                    ),
                    TrialDoneRecord(
                        trial_id="t02",
                        reserved_timestamp=datetime(2025, 6, 28, 19, 42, 0, tzinfo=JST),
                        registered_timestamp=datetime(2025, 6, 28, 19, 44, 0, tzinfo=JST),
                        worker_node_id="w01",
                        worker_node_name="w01",
                        grid_size=100,
                    ),
                    TrialDoneRecord(
                        trial_id="t04",
                        reserved_timestamp=datetime(2025, 6, 28, 19, 43, 0, tzinfo=JST),
                        registered_timestamp=datetime(2025, 6, 28, 19, 45, 0, tzinfo=JST),
                        worker_node_id="w02",
                        worker_node_name="w02",
                        grid_size=50,
                    ),
                ],
                total_grid=1000,
                done_grid=400,
            ),
            StudyProgressSummary(
                study_id="s01",
                study_name="s01",
                total_grid=1000,
                done_grid=400,
                grid_velocity=0.6,
                eta=datetime(2025, 6, 28, 20, 1, 40, tzinfo=JST),
                worker_efficiencies=[
                    WorkerEfficiency(worker_id="w01", worker_name="w01", grid_velocity=0.4),
                    WorkerEfficiency(worker_id="w02", worker_name="w02", grid_velocity=0.2),
                ],
            ),
            id="Normal double worker",
        ),
    ],
)
def test_report_study_progress(
    report_material: ReportMaterial,
    expected: StudyProgressSummary,
) -> None:
    cutoff_sec = 500
    actual = report_study_progress(NOW, cutoff_sec, report_material)
    assert actual.study_id == expected.study_id
    assert actual.study_name == expected.study_name
    assert actual.total_grid == expected.total_grid
    assert actual.done_grid == expected.done_grid
    assert actual.grid_velocity == pytest.approx(expected.grid_velocity)
    assert actual.eta == expected.eta
    assert actual.worker_efficiencies == expected.worker_efficiencies
